from contextvars import ContextVar
from typing import Any

from starlette.requests import Request
from starlette.types import ASGIApp, Receive, Scope, Send

REQUEST_CTX_KEY = "_request"

_request_ctx_var: ContextVar[dict[str, Request]] = ContextVar(REQUEST_CTX_KEY, default=None)


def get_request() -> Request:
    return _request_ctx_var.get()


class AsgiRequestMiddelware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> Any:
        if scope["type"] == "http":
            request = Request(scope)
            _request_ctx_var.set(request)

        await self.app(scope, receive, send)



from uiwiz.element import Element


class Radio(Element):
    root_class: str = "radio"

    def __init__(self, name: str, checked: bool = False) -> None:
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["type"] = "radio"
        if checked:
            self.attributes["checked"] = "checked"
        self.classes()



import markdown2

from uiwiz.elements.html import Html


class Markdown(Html):
    def __init__(self, content="") -> None:
        super().__init__(content=content)
        self.render_html = False
        self.classes("markdown-body")

    def render_self(self, *_) -> str:
        output = ""
        self.content = markdown2.markdown(self.content, extras=["fenced-code-blocks"])
        self.content = self.content.replace("<ul>", '<ul class="list-disc">')
        self.content = self.content.replace("<ol>", '<ol class="list-decimal">')
        self.render_html = True
        output = super().render_self()
        return output



from uiwiz.element import Element


class Range(Element):
    root_class: str = "range "

    def __init__(self, min: int, max: int, value: int, name: str, step: int = None) -> None:
        super().__init__("input")
        self.attributes["type"] = "range"
        self.attributes["name"] = name
        self.attributes["min"] = min
        self.attributes["max"] = max
        self.attributes["value"] = value
        if step:
            self.attributes["step"] = step

        self.classes()

    @property
    def min(self) -> int:
        return self.attributes["min"]

    @min.setter
    def min(self, value: int):
        self.attributes["min"] = value
        return self

    @property
    def max(self) -> int:
        return self.attributes["max"]

    @max.setter
    def max(self, value: int):
        self.attributes["max"] = value
        return self



import itertools
from typing import TypeVar, Union

import pandas as pd
from pydantic import BaseModel

from uiwiz.element import Element
from uiwiz.elements.col import Col
from uiwiz.elements.row import Row
from uiwiz.elements.table import Table


def __display_name__(input: str) -> str:
    return input.replace("_", " ")


T = TypeVar("T")


def show(
    data: Union[BaseModel, T, dict, list[T]],
    container_classes: str = "card w-96 bg-base-100 shadow-lg",
    label_classes: str = "flex-auto w-36 font-bold",
    limit: int = 10,
) -> Element:
    if isinstance(data, list):
        return render_list(data, limit)
    else:
        return render_instance(data, container_classes, label_classes)


def render_list(data: list[T], limit: int) -> Element:
    data_to_render = data if len(data) < limit else data[:limit]
    data_to_render = [vars(d) if not isinstance(d, dict) else d for d in data_to_render]
    return __render_list__(data_to_render)


def render_instance(
    instance: Union[BaseModel, T, dict, list[T]],
    container_classes: str = "card w-96 bg-base-100 shadow-lg",
    label_classes: str = "flex-auto w-36 font-bold",
) -> Element:
    fields: dict
    if isinstance(instance, dict):
        fields = instance
    else:
        fields = vars(instance)

    return __render_data__(fields, container_classes, label_classes)


def __render_list__(data: list[dict]) -> Element:
    return Table(pd.DataFrame(data).rename(columns=__display_name__))


def __render_data__(data: dict, container_classes: str, label_classes: str) -> Element:
    with Element().classes(container_classes) as card:
        with Col():
            for k, v in data.items():
                with Row():
                    Element(content=__display_name__(k)).classes(label_classes)
                    Element(content=v)
    return card



import functools
import inspect
import logging
import os
from html import escape
from mimetypes import guess_type
from pathlib import Path
from typing import Callable, Optional, Union

from fastapi import FastAPI, Response
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.requests import Request

from uiwiz.asgi_request_middelware import AsgiRequestMiddelware
from uiwiz.element import Element
from uiwiz.frame import Frame
from uiwiz.page_route import PageRouter
from uiwiz.shared import resources
from uiwiz.static_middelware import AsgiTtlMiddelware

logger = logging.getLogger("uiwiz")
logger.addHandler(logging.NullHandler())


class UiwizApp(FastAPI):
    def __init__(
        self,
        toast_delay: int = 2500,
        error_classes: str = "alert alert-error",
        cache_age: int = 14400,
        theme: Optional[str] = None,
        auth_header: Optional[str] = None,
        title: Optional[str] = "UiWiz",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.toast_delay = toast_delay
        self.error_classes = error_classes
        if theme:
            self.theme = f"data-theme={theme}"
        else:
            self.theme = theme
        self.auth_header = auth_header
        self.title = title
        self.templates = Jinja2Templates(Path(__file__).parent / "templates")
        self.add_static_files("/static", Path(__file__).parent / "static")

        self.add_middleware(AsgiRequestMiddelware)
        self.add_middleware(GZipMiddleware)
        self.add_middleware(AsgiTtlMiddelware, cache_age=cache_age)
        self.extensions: dict[str, Path] = {}
        self.app_paths: dict[str, Path] = {}

        @self.get("/_static/default.js")
        def return_default_js(request: Request):
            return self.render(request, template_name="default.js", media_type="application/javascript")

        @self.exception_handler(RequestValidationError)
        async def handle_validation_error(request: Request, exc: RequestValidationError):
            fields_with_errors = [item.get("loc")[1] for item in exc.errors() if item.get("loc")[1] in exc.body]
            ok_fields = [item for item in exc.body.keys() if item not in fields_with_errors]
            return JSONResponse(
                status_code=422,
                content=jsonable_encoder(
                    {"detail": exc.errors(), "fieldErrors": fields_with_errors, "fieldOk": ok_fields}
                ),
            )

        @self.get("/_static/extension/{extension}/{filename}")
        def get_extension(extension: str, filename: str):
            resource_key = f"{extension}/{filename}"
            if resource_key not in resources:
                return Response(status_code=404)

            with open(resources[resource_key]) as f:
                content = f.read()

            content_type, _ = guess_type(resource_key)
            return Response(content, media_type=content_type)

    def __get_title__(self, frame: Frame, route_title: Optional[str] = None) -> str:
        title = self.title

        if route_title:
            title = route_title

        if dynamic_title := frame.title:
            title = dynamic_title

        return title

    def render(
        self,
        request: Request,
        title: Optional[str] = None,
        status_code: int = 200,
        template_name: str = "default.html",
        media_type: Optional[str] = None,
    ):
        frame = Frame.get_stack()
        html = frame.render()
        ext = frame.render_ext()
        page_title = self.__get_title__(frame, title)
        theme = self.theme
        if cookie_theme := request.cookies.get("data-theme"):
            theme = f"data-theme={escape(cookie_theme)}"
        return self.return_funtion_response(
            self.templates.TemplateResponse(
                name=template_name,
                context={
                    "request": request,
                    "root_element": [html],
                    "title": page_title,
                    "theme": theme,
                    "ext": ext,
                    "toast_delay": self.toast_delay,
                    "error_classes": self.error_classes,
                    "auth_header_name": self.auth_header,
                    "description_content": frame.meta_description_content,
                },
                status_code=status_code,
                headers={"Cache-Control": "no-store", "X-uiwiz-Content": "page"},
                media_type=media_type,
            )
        )

    def render_api(self, status_code: int = 200):
        return self.return_funtion_response(
            HTMLResponse(
                Frame.get_stack().render(),
                status_code,
                {"Cache-Control": "no-store", "X-uiwiz-Content": "partial-ui"},
            )
        )

    def route_exists(self, path: str) -> None:
        return path in list(self.app_paths.values())

    def remove_route(self, path: str) -> None:
        """Remove routes with the given path."""
        self.routes[:] = [r for r in self.routes if getattr(r, "path", None) != path]

    def add_static_files(self, url_path: str, local_directory: Union[str, Path]) -> None:
        self.mount(url_path, StaticFiles(directory=str(local_directory)))

    def post(self, path: str, *args, **kwargs):
        s = super()

        @functools.wraps(s.post)
        def decorator(func: Callable, *args, **kwargs) -> Callable:
            if not self.route_exists(path):
                self.app_paths[func] = path
            return s.post(path, *args, **kwargs)(func)

        return decorator

    def page(
        self,
        path: str,
        *args,
        title: Optional[str] = None,
        favicon: Optional[str] = None,
    ) -> Callable:
        def decorator(func: Callable, *args, **kwargs) -> Callable:
            self.remove_route(path)
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs) -> Response:
                # Create frame before function is called
                Frame.get_stack()
                Element().classes("flex flex-col min-h-screen h-full")
                request = dec_kwargs["request"]
                # NOTE cleaning up the keyword args so the signature is consistent with "func" again
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result
                if isinstance(result, Response):
                    return self.return_funtion_response(result)

                return self.render(request, title)

            params = [p for p in inspect.signature(func).parameters.values()]
            if "request" not in {p.name for p in params}:
                request = inspect.Parameter(
                    "request",
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=Request,
                )
                params.insert(0, request)
            decorated.__signature__ = inspect.Signature(params)

            if not self.route_exists(path):
                self.app_paths[decorated] = path

            return self.get(path)(decorated)

        return decorator

    def ui(self, path: str) -> Callable:
        def decorator(func: Callable) -> Callable:
            self.remove_route(path)
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs) -> Response:
                # Create frame before function is called
                Frame.get_stack()
                # NOTE cleaning up the keyword args so the signature is consistent with "func" again
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result
                if isinstance(result, Response):  # NOTE if setup returns a response, we don't need to render the page
                    return self.return_funtion_response(result)

                return self.render_api()

            request = inspect.Parameter("request", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=Request)
            params = [p for p in inspect.signature(func).parameters.values()]
            for p in params:
                if p.annotation == inspect.Signature.empty:
                    p._annotation = Request
            if "request" not in {p.name for p in params}:
                params.insert(0, request)
            decorated.__signature__ = inspect.Signature(params)

            if not self.route_exists(path):
                self.app_paths[decorated] = path

            return self.post(path)(decorated)

        return decorator

    def add_page_router(self, page_router: PageRouter):
        for key, value in page_router.paths.items():
            if not self.route_exists(key):
                self.app_paths[value.get("func")] = key
            type = value.get("type")
            if type == "page":
                self.page(key, *value.get("args"), **value.get("kwargs"))(value.get("func"))
            if type == "ui":
                self.ui(key)(value.get("func"))

    def return_funtion_response(self, response: Response) -> Response:
        Frame.get_stack().del_stack()
        return response



from typing import Any, Callable, List, Literal, TypedDict


class Path(TypedDict):
    func: Callable
    type: Literal["page", "ui"]
    args: Any
    kwargs: Any


class PageRouter:
    def __init__(self) -> None:
        self.paths: dict[str, Path] = {}

    def page(
        self,
        path: str,
        *args,
        **kwargs,
    ) -> Callable:
        def setup(func):
            self.paths[path] = {"func": func, "type": "page", "args": args, "kwargs": kwargs}
            return func

        return setup

    def ui(self, path: str, *args, **kwargs) -> Callable:
        def setup(func):
            self.paths[path] = {"func": func, "type": "ui", "args": args, "kwargs": kwargs}
            return func

        return setup



from typing import Iterable, Union

from uiwiz.element import Element


class Dict(Element):
    def __init__(self, data: Union[Iterable[dict], dict]) -> None:
        if not data:
            raise ValueError("Data cannot be None or empty")
        if not isinstance(data, (Iterable, dict)):
            raise ValueError("Data not of type list or dict")
        super().__init__()
        self.generate(data)

    def generate(self, data):
        def format_json(obj):
            if isinstance(obj, dict):
                with Element().classes("card w-96 bg-base-100 shadow-md w-full mb-5"):
                    for key, value in obj.items():
                        with Element().classes("flex flex-row flex-wrap gap-2"):  # row
                            key_element = Element()
                            if isinstance(value, (dict, list)):
                                with key_element.classes("collapse collapse-arrow pl-2"):
                                    ch = Element("input")
                                    ch.attributes["type"] = "checkbox"
                                    ch.attributes["checked"] = "checked"
                                    Element(content=f"{key}:").classes("collapse-title text-sm min-h-8 p-0")
                                    with Element().classes("collapse-content"):
                                        format_json(value)
                            else:
                                key_element.content = f"{key}:"
                                key_element.classes("pl-2")
                                Element("p", value).classes("text-primary pl-2")
            elif isinstance(obj, list):
                for item in obj:
                    if isinstance(item, dict):
                        format_json(item)
                    else:
                        Element(content=item)
            else:
                Element(content=obj).classes("text-primary")

        format_json(data)



from datetime import date, datetime
from typing import Optional, Union

from uiwiz.element import Element


class Datepicker(Element):
    def __init__(self, name: str, value: Optional[datetime] = None) -> None:
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["type"] = "date"
        if value:
            self.default_date(value)

    def default_date(self, date: Union[datetime, date]) -> "Datepicker":
        if date is None:
            raise ValueError("Date cannot be None")
        self.value = date.strftime("%Y-%m-%d")
        return self



from typing import Callable, Union

from uiwiz.element import Element


class Upload(Element):
    root_class: str = "file-input "
    _classes: str = "file-input-bordered file-input-sm"

    def __init__(
        self,
        name: str,
        on_upload: Callable,
        target: Union[Callable, str, Element] = None,
        trigger: str = "change",
        swap: str = None,
    ) -> "Upload":
        super().__init__("input")
        self.attributes["type"] = "file"
        self.attributes["name"] = name
        self.classes(Upload._classes)

        self.event = {
            "func": on_upload,
            "trigger": trigger,
            "target": target,
            "swap": swap,
            "hx-encoding": "multipart/form-data",
        }



import inspect
from dataclasses import dataclass
from datetime import date, datetime
from typing import Annotated, Literal, Optional, Tuple, Union, get_args, get_origin, get_type_hints

from pydantic import BaseModel
from pydantic_core import PydanticUndefinedType

from uiwiz.element import Element
from uiwiz.elements.button import Button
from uiwiz.elements.checkbox import Checkbox
from uiwiz.elements.datepicker import Datepicker
from uiwiz.elements.divider import Divider
from uiwiz.elements.dropdown import Dropdown
from uiwiz.elements.form import Form
from uiwiz.elements.hidden_input import HiddenInput
from uiwiz.elements.input import Input
from uiwiz.elements.label import Label
from uiwiz.elements.radio import Radio
from uiwiz.elements.textarea import TextArea
from uiwiz.elements.toggle import Toggle
from uiwiz.show import __display_name__


@dataclass
class UiAnno:
    type: Union[Input, TextArea, Checkbox] = None
    placeholder: Optional[str] = None
    classes: Optional[str] = None


switch = {
    str: Input,
    int: Input,
    float: Input,
    bool: Toggle,
    Literal: Dropdown,
    date: Datepicker,
    datetime: Datepicker,
}


class ModelForm:
    model: BaseModel
    compact: bool
    form: Form
    button: Button
    label_classes: str
    instance: Optional[BaseModel]

    def __init__(
        self,
        model: BaseModel,
        compact: bool = True,
        card_classes: str = "card w-full bg-base-100 shadow-lg",
        label_classes: str = "flex-auto w-52",
        **kwargs,  # override fields with custom ui
    ):
        self.model = model
        self.instance = None

        if isinstance(model, BaseModel):
            self.model = model.__class__
            self.instance = model
        self.compact = compact
        self.label_classes = label_classes
        self.card_classes = card_classes
        self.render_model(**kwargs)
        self.button.render_html = False

    def on_submit(self, *args, **kwargs) -> "ModelForm":
        self.button.render_html = True
        self.form.on_submit(*args, **kwargs)
        return self

    def render_model(self, **kwargs) -> Form:
        if issubclass(self.model, BaseModel) == False:
            raise ValueError("type must be a pydantic model")

        with Form().classes(self.card_classes) as form:
            hints = get_type_hints(self.model, include_extras=True)
            for key, field_type in hints.items():
                args = get_args(field_type)
                annotated = Annotated == get_origin(field_type)
                if key in kwargs:
                    self.render_key_override(args, key, field_type, **kwargs)
                else:
                    self.render_type_hint_without_args(args, annotated, field_type, key)
                    self.render_with_args_annotated(args, annotated, field_type, key)
            self.button = Button("Save")
        self.form = form

    def render_key_override(self, args: Tuple, key: str, field_type: type, **kwargs) -> None:
        if key in kwargs:
            model_args: dict = kwargs[key]
            model = model_args.pop("ui")
            placeholder = str(model_args.pop("placeholder", key))
            if not issubclass(model, Element):
                raise ValueError("type must be type of Element")
            if "name" not in model_args:
                model_args["name"] = key

            _, arg = self.get_type_and_uianno(args)
            classes = None
            if arg:
                classes = arg.classes
            self.render_element(model, key, classes, placeholder=placeholder, **model_args)
        else:
            raise ValueError("key not found in kwargs. Unable to render")

    def render_type_hint_without_args(self, args: Tuple, annotated: bool, field_type, key) -> Element:
        if len(args) == 0:
            if annotated:
                self.render_element(switch.get(field_type), key, placeholder=key)

            self.render_element(switch.get(field_type), key, placeholder=key)

    def render_with_args_annotated(self, args: Tuple, annotated: bool, field_type: Tuple, key: str) -> Element:
        if len(args) > 0:
            if annotated:
                field_type, ele = self.get_type_and_uianno(args)
                if ele:
                    placeholder = ele.placeholder if ele.placeholder else key
                    if field_args := get_args(field_type):
                        self.render_element_radio(ele, key, field_args)
                    else:
                        self.render_element(ele.type, key, ele.classes, placeholder=placeholder)
            else:
                ele = switch.get(get_origin(field_type))
                if ele:
                    self.render_element_dropdown(field_type, key, self.model.model_fields[key].default)

    def get_type_and_uianno(self, args: Tuple) -> Optional[Tuple[type, Optional[UiAnno]]]:
        if args:
            field_type = args[0]
            for arg in args:
                if isinstance(arg, UiAnno):
                    return field_type, arg
            return field_type, None
        return None, None

    def render_element(
        self,
        ele: Element,
        key: str,
        classes: Optional[str] = None,
        field_arg: Optional[str] = None,
        **kwargs,
    ) -> None:
        kwargs = {**{"name": key}, **kwargs}
        placeholder = "placeholder"
        kwargs[placeholder] = __display_name__(kwargs[placeholder])
        compact = self.compact
        with Element().classes("flex flex-nowrap w-full"):
            ele_args = [item[0] for item in inspect.signature(ele.__init__).parameters.items()]
            compact = self.extend_kwargs(kwargs, ele_args, key, field_arg, ele)

            label: Optional[Label] = None
            if ele is not HiddenInput:
                label = Label(kwargs[placeholder])

                if compact and inspect.signature(ele).parameters.get(placeholder):
                    label.render_html = False
                else:
                    label.classes(self.label_classes)
                    label.render_html = True
            if not inspect.signature(ele).parameters.get(placeholder):
                kwargs.pop(placeholder)

            if self.instance and "placeholder" in ele_args:
                kwargs["placeholder"] = getattr(self.instance, key)
            el: Element = ele(**kwargs)
            if classes:
                el.classes(classes)
            if label:
                label.set_for(el)

    def extend_kwargs(self, kwargs: dict, ele_args: list[str], key: str, field_arg: str, ele: Element) -> bool:
        compact = self.compact
        if self.instance and "value" in ele_args:
            kwargs["value"] = getattr(self.instance, key)
            compact = False

        # radio button
        if self.instance and getattr(self.instance, key) == field_arg:
            kwargs["checked"] = "checked"
            compact = False
        elif self.instance and ele == Toggle:
            kwargs["checked"] = "checked"
            compact = False
        return compact

    def render_element_radio(
        self,
        ele: UiAnno,
        key: str,
        field_args: Tuple,
    ) -> None:
        Divider()
        for arg in field_args:
            self.render_element(Radio, key, ele.classes, placeholder=arg, field_arg=arg)
        Divider()

    def render_element_dropdown(self, field_class: type, key: str, placeholder: Optional[str]) -> None:
        if isinstance(placeholder, PydanticUndefinedType):
            placeholder = __display_name__(key)

        self.render_element(Dropdown, key, placeholder=placeholder, items=get_args(field_class))



from uiwiz.element import Element


class Nav(Element):
    root_class: str = "navbar w-full "
    _classes: str = "bg-neutral"

    def __init__(self) -> None:
        super().__init__()
        self.classes(Nav._classes)
